package com.persagy.bdtp.adm.service;

import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.persagy.bdtp.adm.entity.db.*;

import java.util.List;

public interface IConfigService {

	List<AdmM2dEquip> queryM2dEquip(String projectId);

	List<AdmInfosConfig> queryInfosConfig(String projectId);

	List<AdmRelsConfig> queryRelsConfig(String projectId);

	List<AdmPipeConfig> queryPipeConfig(String projectId);

	List<AdmConfig> queryCommonConfig(String projectId);

	List<AdmComponent> queryComponent(String projectId);

	List<AdmContainerConfig> queryContainerConfig(String projectId);

	ArrayNode queryNewEquipTypes(String projectId);

	ObjectNode queryRenameEquipTypes(String projectId);

    List<AdmDefineProblemType> queryDefineProblemTypes();

}
